/*
 *  main.c    Main NC9100
 *  2003-11-10
 */

#include "at91lib.h"
#include "includes.h"
#include "string.h"
#include "at91uart.h"
#include "setup.h"
#include "log.h"
#include "test.h"
#include "config.h"
#include "lwip/sockets.h"
#include "flash.h"
#include "crc16.h"
#include "timer.h"
#include "atc.h"
#include "arch/lib.h" 
#include "arch/sys_arch.h"
#include "lwip/tcpip.h"
#include "lwip/memp.h"
#include "lwip/pbuf.h"
#include "lwip/netif.h"
#include "lwip/dhcp.h"
#include "webs/fs.h"
#include "webs/httpd.h"
#include "telnetd.h"
#include "download.h"
#include "digiod.h"
#include "lwip/stats.h"
#include "dump.h"   
#include "et9000if.h"
 
#define SERIAL_TASK1_ID    2132
#define SERIAL_TASK2_ID    2133
#define NET_TASK1_ID       2134
#define NET_TASK2_ID       2135
#define SERVER_TASK1_ID    2136
#define SERVER_TASK2_ID    2137


#define STACKSIZE 1024
#define NO_CHAR   1024
#define BG_TASKS 3
#define TIMER_TASK_ID 2141

unsigned int initStack[STACKSIZE];
unsigned int serialReceive1Stack[STACKSIZE];
unsigned int serialReceive2Stack[STACKSIZE];
unsigned int netReceive1Stack[STACKSIZE];
unsigned int netReceive2Stack[STACKSIZE];
unsigned int server1Stack[STACKSIZE];
unsigned int server2Stack[STACKSIZE];
unsigned int timerStack[STACKSIZE];

int logLevel, flash, test;
int atData;
char *pversion = "Net Controller STD 0.10";
char id;
INT32U my_local_ip;
INT16U my_serial1;
INT16U my_serial2;

struct netif network;

int socket1;
int socket2;
int socket1_srv;
int socket2_srv;

struct sockaddr_in udp_addr1, udp_addr2;

extern PARM Set;
extern SYS Sets;
extern struct atPort_t atPort1, atPort2;
extern struct taskdump_ taskdump;

OS_EVENT *startReceive1;
OS_EVENT *startAccept1;
OS_EVENT *startReceive2;
OS_EVENT *startAccept2;

char sostream[1000];
char cardId(void);

void SetAddr (struct sockaddr_in *addr, u32_t ip, u16_t tcp);

/*
 *  SerialReceive1Task()
 *  Serial Receive Port-1
 */
void SerialReceive1Task(void *i)
{
  char buffer1[NO_CHAR];
  int readByte;
  int ret;
  struct sockaddr_in my_addr;
  struct sockaddr_in his_addr;
  u32_t ip;
      
  while (1) {
    taskdump.serial1++;
    ret = serialRead(SERIAL_1, buffer1, NO_CHAR, &readByte, 0);
    Log(4, "Serial-1 Recv ", Z4, &readByte);

    if(ret == AT91_ERR_DATA)            // Receive frame or parity error ?
      Log(1,"Port-1 Err Recv", 0, 0);
    
    if(ret == AT91_ERR_RELEASED) {      // Released ?
      serialIoCtl(SERIAL_1, 'B', 9600);
      serialIoCtl(SERIAL_1, 'D', SO_DB8);
      SetupSerial(LP_1);
      continue;
    }
    
    if(socket1 == -1) {                // Not connected ?
      if(!(Set.sopt1 & SO_LC))         // Local connect not enabled
        continue; 
      if(Set.sopt1 & SO_ATC) {         // AT mode ?
        if(!(ip = AtCommand(&atPort1, buffer1, readByte)))    // AT Command ?
          continue;
        readByte = 0;                     // Valid IP adress
      }    
      else                                // Direct mode ?
        ip = Set.remote_ip1;
      
      if(Set.option & OP_UDP)           // UDP ?
        socket1 = socket(AF_INET, SOCK_DGRAM, 0);
      else                              // TCP ?
        socket1 = socket(AF_INET, SOCK_STREAM, 0);

      SetAddr (&my_addr, INADDR_ANY, Set.local_tcp1);
      SetAddr (&his_addr, Set.remote_ip1, Set.remote_tcp1);
      bind(socket1, (struct sockaddr *)&my_addr, sizeof(my_addr));
      if(connect(socket1, (struct sockaddr *)&his_addr, sizeof(his_addr)) < 0){
        close(socket1);
        socket1 = -1;
        AtMode(&atPort1, 4);
        Log(1,"Port-1 Err connect", 0, 0);
        continue;
      }
      AtMode(&atPort1, 1);
      Log(2,"Port-1 Opened", 0, 0);
      SetTimer(SERIAL_1, Set.connect);
      SetTimer(ACTIVITY, Set.activity);
      OSSemPost(startReceive1);

      if (!readByte)                    // Nothing to send ?
        continue;
        
    } // Not connected  
    
    if(socket1 != -1) {                 // Connected ? 
      if(Set.sopt1 & SO_ATC) {           // AT mode ?
        if(!(ip = AtCommand(&atPort1, buffer1, readByte)))    // AT Command ?
          continue;
      }    
      if(Set.option & OP_UDP) {         // UDP ? 
        ret = sendto(socket1, buffer1, readByte, 0,(struct sockaddr *)&udp_addr1, sizeof(udp_addr1));
        if(ret < 0)
          Log(1,"Port-1 Err send", Z4, &ret);
        else
          Log(2,"Port-1 Send", Z4, &ret);
        SetTimer(SERIAL_1, Set.connect);
        SetTimer(ACTIVITY, Set.activity);
      }
      else {                            // TCP ?
        ret = send(socket1, buffer1, readByte, 0);
        if (ret < 0)
          Log(1,"Port-1 Err send", Z4, &ret);
        else
          Log(2,"Port-1 Send", Z4, &ret);
        SetTimer(SERIAL_1, Set.connect);
        SetTimer(ACTIVITY, Set.activity);
      }
    }
  }
}

/*
 *  SerialReceive2Task()
 *  Serial Receive Port-2
 */
void SerialReceive2Task(void *i)
{
  char buffer2[NO_CHAR];
  int readByte;
  int ret;
  struct sockaddr_in my_addr;
  struct sockaddr_in his_addr;
  u32_t ip;
      
  while (1) {
    taskdump.serial2++;
    ret = serialRead(SERIAL_2, buffer2, NO_CHAR, &readByte, 0);
    Log(4, "Serial-2 Recv ", Z4, &readByte);
  

    if(ret == AT91_ERR_DATA)            // Receive frame or parity error ?
      Log(1,"Port-2 Err Recv", 0, 0);

    if(ret == AT91_ERR_RELEASED) {      // Released ?
      serialIoCtl(SERIAL_2, 'B', 9600);
      serialIoCtl(SERIAL_2, 'D', SO_DB8);
      SetupSerial(LP_2);
      continue;
    }

    if(socket2 == -1) {                // Not connected ? 
      if(!(Set.sopt2 & SO_LC))         // Local connect not enabled
        continue; 
      if(Set.proto2 & SO_ATC) {           // AT mode ?
        if(!(ip = AtCommand(&atPort2, buffer2, readByte)))    // AT Command ?
          continue;
        readByte = 0;                     // Valid IP adress
      }    
      else                                // Direct mode ?
        ip = Set.remote_ip2;
      
      if(Set.option & OP_UDP)           // UDP ?
        socket2 = socket(AF_INET, SOCK_DGRAM, 0);
      else                              // TCP ?
        socket2 = socket(AF_INET, SOCK_STREAM, 0);

      SetAddr (&my_addr, INADDR_ANY, Set.local_tcp2);
      SetAddr (&his_addr, Set.remote_ip2, Set.remote_tcp2);
      bind(socket2, (struct sockaddr *)&my_addr, sizeof(my_addr));

      if(connect(socket2, (struct sockaddr *)&his_addr, sizeof(his_addr)) < 0){
        close(socket2);
        socket2 = -1;
        AtMode(&atPort2, 4);
        Log(1,"Port-2 Err connect", 0, 0);
        continue;
      }
      AtMode(&atPort2, 1);
      Log(2,"Port-2 Opened", 0, 0);
      SetTimer(SERIAL_2, Set.connect);
      SetTimer(ACTIVITY, Set.activity);
      OSSemPost(startReceive2);

      if (!readByte)                    // Nothing to send ?
        continue;
        
    } // Not connected  

    if(socket2 != -1) {                 // Connected ? 
      if(Set.sopt2 & SO_ATC) {           // AT mode ?
        if(!(ip = AtCommand(&atPort2, buffer2, readByte)))    // AT Command ?
          continue;
      }    
      if(Set.option & OP_UDP) {         // UDP ?
        ret = readByte = sendto(socket2, buffer2, readByte, 0,(struct sockaddr *)&udp_addr2, sizeof(udp_addr2));
        if (ret < 0)
          Log(1,"Port-2 Err send", Z4, &ret);
        else
          Log(2,"Port-2 Send", Z4, &ret);
        SetTimer(SERIAL_2, Set.connect);
        SetTimer(ACTIVITY, Set.activity);
      }
      else {                            // TCP ?
        ret = send(socket2, buffer2, readByte, 0);
        if (ret < 0)
          Log(1,"Port-2 Err send", Z4, &ret);
        else
          Log(2,"Port-2 Send", Z4, &ret);
        SetTimer(SERIAL_2, Set.connect);
        SetTimer(ACTIVITY, Set.activity);
      }
    }
  }
}

/*
 *  NetReceive1Task()
 *  Net Receive Socket-1
 */
void NetReceive1Task(void *i)
{
  char buf[500];                     // ska vara [100] 
  int len;
  INT8U err;
  int fromlen;
  struct sockaddr_in his_addr;
  
  startReceive1 = OSSemCreate(0); 
  
  while (1) {
    taskdump.net1++;
    OSSemPend(startReceive1, 0, &err);     // Wait for start of receive after connect

    SetAddr (&his_addr, Set.remote_ip1, Set.local_tcp1);
    if(Set.option & OP_UDP) {        // UDP ? 
  		fromlen = sizeof(udp_addr1);
      while ((len = recvfrom(socket1, buf, sizeof(buf), 0, (struct sockaddr *)&udp_addr1, &fromlen)) > 0) {
        Log(2,"Port-1 Received", Z4, &len);
        serialWrite(SERIAL_1, buf, len);
        SetTimer(SERIAL_1, Set.connect);
        SetTimer(ACTIVITY, Set.activity);
      }
    }
    else {                          // TCP ?                  
    	while ((len = recv(socket1, buf, sizeof(buf), 0)) > 0) {
        Log(2, "socket = ", Z4, &socket1);
        Log(2,"Port-1 Received", Z4, &len);
        serialWrite(SERIAL_1, buf, len);
        SetTimer(SERIAL_1, Set.connect);
        SetTimer(ACTIVITY, Set.activity);
      }
    }
    close(socket1);
    socket1 = -1;
    SetAddr (&udp_addr1, Set.remote_ip1, Set.remote_tcp1);
    AtMode(&atPort1, 3);
    if(!(Set.sopt1 & SO_ATC)) {
      serialIoCtl(SERIAL_1, 'A', 0);    // DCD off
      SetTimer(DCD_1, 2);
    }
    Log(2,"Port-1 Closed", 0, 0);
    OSSemPost(startAccept1);
  }
}

/*
 *  NetReceive2Task()
 *  Net Receive Socket-2
 */
void NetReceive2Task(void *i)
{
  char buf[500];                     // ska vara [100] 
  INT16U len;
  INT8U err;
  int fromlen;
  struct sockaddr_in his_addr;
  
  startReceive2 = OSSemCreate(0); 
  
  while (1) {
    taskdump.net2++;
    OSSemPend(startReceive2, 0, &err);     // Wait for start of receive after connect

    SetAddr (&his_addr, Set.remote_ip2, Set.local_tcp2);
    if(Set.option & OP_UDP) {         // UDP ?
   		fromlen = sizeof(udp_addr2);
      while ((len = recvfrom(socket2, buf, sizeof(buf), 0, (struct sockaddr *)&udp_addr2, &fromlen)) > 0) {
        Log(2,"Port-2 Received", Z4, &len);
        serialWrite(SERIAL_2, buf, len);
        SetTimer(SERIAL_2, Set.connect);
        SetTimer(ACTIVITY, Set.activity);
      }
    }
    else {                            // TCP ?                
    	while ((len = recv(socket2, buf, sizeof(buf), 0)) > 0) {
        Log(2,"Port-2 Received", Z4, &len);
        serialWrite(SERIAL_2, buf, len);
        SetTimer(SERIAL_2, Set.connect);
        SetTimer(ACTIVITY, Set.activity);
      }
    }

    close(socket2);
    socket2 = -1;
    SetAddr (&udp_addr2, Set.remote_ip2, Set.remote_tcp2);
    AtMode(&atPort2, 1);
    if(!(Set.sopt2 & SO_ATC)) {
      serialIoCtl(SERIAL_2, 'A', 0);    // DCD off
      SetTimer(DCD_2, 2);
    }
    Log(2,"Port-2 Closed", 0, 0);
    OSSemPost(startAccept2);
  }
}

/*
 *  Server1Task()
 *  Server listner socket-1
 */
void Server1Task(void *i)
{
  struct sockaddr_in his_addr, my_addr;
  int fromlen;
  INT8U err;

  startAccept1 = OSSemCreate(0); 

  if(Set.option & OP_UDP)             // UDP ?
    socket1_srv = socket(AF_INET, SOCK_DGRAM, 0);
  else                                // TCP ?
    socket1_srv = socket(AF_INET, SOCK_STREAM, 0);

  SetAddr (&his_addr, Set.remote_ip1, 0);
  SetAddr (&my_addr, my_local_ip, Set.remote_tcp1);
  bind(socket1_srv, (struct sockaddr *)&my_addr, sizeof(my_addr));

  if(Set.option & OP_UDP)             // UDP ? 
    Log(2,"Port-1 UDP", D5, &Set.local_tcp1);
  else {                              // TCP ?
    listen(socket1_srv, 5);						// allow a queue of 5
    Log(2,"Port-1 Closed", D5, &Set.local_tcp1);
  } 

  while (1) {
    taskdump.server1++;

    fromlen = sizeof(his_addr);

    if(Set.option & OP_UDP) {         // UDP ? 
      SetAddr (&udp_addr1, Set.remote_ip1, Set.remote_tcp1);
      socket1 = socket1_srv;
      OSSemPost(startReceive1);
    }
    else {                            // TCP ?
      if((socket1 = accept(socket1_srv, (struct sockaddr * )&his_addr, &fromlen)) < 0){
        Log(1,"Server accept error", 0, 0);
        continue;
      }
      
      if(Set.option & OP_PRM) {       // Permissions
        if(Set.remote_ip1 != htonl(his_addr.sin_addr.s_addr)) {
          close(socket1);
          continue;
        }
      }
      
      SetTimer(SERIAL_1, Set.connect);
      SetTimer(ACTIVITY, Set.activity);
      AtMode(&atPort1, 1);
      Log(2,"Port-1 Opened", 0, 0);
      OSSemPost(startReceive1);
    }
    OSSemPend(startAccept1, 0, &err);
  }
}

/*
 *  Server2Task()
 *  Server Listen Socket-2
 */
void Server2Task(void *i)
{
  struct sockaddr_in his_addr, my_addr;
  int fromlen;
  INT8U err;

  startAccept2 = OSSemCreate(0); 

  if(Set.option & OP_UDP)             // UDP ?
    socket2_srv = socket(AF_INET, SOCK_DGRAM, 0);
  else                               // TCP ?
    socket2_srv = socket(AF_INET, SOCK_STREAM, 0);

  SetAddr (&his_addr, Set.remote_ip2, 0);
  SetAddr (&my_addr, my_local_ip, Set.remote_tcp2);
  bind(socket2_srv, (struct sockaddr *)&my_addr, sizeof(my_addr));

  if(Set.option & OP_UDP)            // TCP ? 
    Log(2,"Port-2 UDP", D5, &Set.local_tcp2);
  else {
    listen(socket2_srv, 5);											// allow a queue of 5 
    Log(2,"Port-2 Closed", D5, &Set.local_tcp2);
  }
  while (1) {
    taskdump.server2++;

    fromlen = sizeof(his_addr);

    if(Set.option & OP_UDP) {         // UDP ? 
      SetAddr (&udp_addr2, Set.remote_ip2, Set.remote_tcp2);
      socket2 = socket2_srv;
      OSSemPost(startReceive2);
    }
    else {                            // TCP ?
      if((socket2 = accept(socket2_srv, (struct sockaddr * )&his_addr, &fromlen)) < 0){
        Log(1,"Server accept error", 0, 0);
        continue;
      }
      
      if(Set.option & OP_PRM) {       // Permissions
        if(Set.remote_ip2 != htonl(his_addr.sin_addr.s_addr)) {
          close(socket2);
          continue;
        }
      }

      SetTimer(SERIAL_2, Set.connect);
      SetTimer(ACTIVITY, Set.activity);
      AtMode(&atPort2, 1);
      Log(2,"Port-2 Opened", 0, 0);
      OSSemPost(startReceive2);
    }
    OSSemPend(startAccept2, 0, &err);
  }
}

static void tcpip_init_done(void *arg)
{
  sys_sem_t *sem;
  sem = arg;
  sys_sem_signal(*sem);
}

void InitTask(void *i)
{
  char temp[15];
  sys_sem_t sem;
  u32_t ipaddr, netmask, gateway, value;
  struct ip_addr dhcp_ip, dhcp_mask, dhcp_gw;
    
  at91OSTimerInit();
  OSStatInit ();
  
  taskdump_init();

  serialInit(SERIAL_1, my_serial1 * 4, Set.sopt1, Set.recv1);
  serialInit(SERIAL_2, my_serial2 * 4, Set.sopt2, Set.recv2);
  
  if(!(Set.sopt1 & SO_ATC))
    serialIoCtl(SERIAL_1, 'A', 1);    // DCD-1 on

  if(!(Set.sopt2 & SO_ATC))
    serialIoCtl(SERIAL_2, 'A', 1);    // DCD-2 on

  LogInit();

  Log (2,pversion,0,0);    
    
  Log (3,"Ethernet", S1, eth_ntoa(Sets.mac,temp));

  value = Set.serial1 * 4;
  Log (3,"Serial-1", D5, &value);
  
  value = Set.serial2 * 4;
  Log (3,"Serial-2", D5, &value);

  if (!test)
    Log (1,"Test error", Z2, &test);

  OSTaskCreateExt(TimerTask, (void *)0, &timerStack[STACKSIZE - 1], 15, TIMER_TASK_ID, &timerStack[0], STACKSIZE, (void *)0, 0x3);

/* Old OSTaskCreate
  OSTaskCreate(TimerTask, (void *)123, &timerStack[STACKSIZE - 1], 15);
*/

  AtInit();

  stats_init();
  sys_init();
  mem_init();
  memp_init();
  pbuf_init();
  
  netif_init();
  
  sem = sys_sem_new(0);
  tcpip_init(tcpip_init_done, &sem);
  sys_sem_wait(sem);
  sys_sem_free(sem);
  
  if(!my_local_ip){
    IP4_ADDR(&dhcp_gw, 0,0,0,0);        
    IP4_ADDR(&dhcp_mask, 0,0,0,0);        
    IP4_ADDR(&dhcp_ip, 0,0,0,0);        
  
    netif_set_default(netif_add(&network, &dhcp_ip, &dhcp_mask, &dhcp_gw, NULL, ethernetif_init, tcpip_input));
    OSTimeDly(500);       // Delay to wait for the ethernet link to come up
    dhcp_start(&network);
  }
  else {
    ipaddr = htonl(my_local_ip);
    netmask = htonl(Set.netmask);
    gateway = htonl(Set.gateway);
  
    netif_set_default(netif_add(&network, (struct ip_addr *)&ipaddr,(struct ip_addr *)&netmask, (struct ip_addr *)&gateway, NULL, ethernetif_init, tcpip_input));
  }  

  socket1 = -1;
  socket2 = -1;
/*
  Arg: 1  Task
       2  No arguments passed to Task()
       3  Set Top-Of-Stack
       4  Priority
       5  Unique ID
       6  Set Bottom-Of-Stack
       7  Stack size
       8  No TCB extension
       9  Enable stack checking OS_TASK_OPT_STK_CHK + OS_TASK_OPT_STK_CLR
*/

  OSTaskCreateExt(SerialReceive1Task, (void *)0, &serialReceive1Stack[STACKSIZE - 1], 9, SERIAL_TASK1_ID, &serialReceive1Stack[0], STACKSIZE, (void *)0, 0x3);
  OSTaskCreateExt(SerialReceive2Task, (void *)0, &serialReceive2Stack[STACKSIZE - 1], 10, SERIAL_TASK2_ID, &serialReceive2Stack[0], STACKSIZE, (void *)0, 0x3);
  OSTaskCreateExt(NetReceive1Task, (void *)0, &netReceive1Stack[STACKSIZE - 1], 11, NET_TASK1_ID, &netReceive1Stack[0], STACKSIZE, (void *)0, 0x3);
  OSTaskCreateExt(NetReceive2Task, (void *)0, &netReceive2Stack[STACKSIZE - 1], 12, NET_TASK2_ID, &netReceive2Stack[0], STACKSIZE, (void *)0, 0x3);

  if(Set.sopt1 & SO_RC)    // remote connect
    OSTaskCreateExt(Server1Task, (void *)0, &server1Stack[STACKSIZE - 1], 13, SERVER_TASK1_ID, &server1Stack[0], STACKSIZE, (void *)0, 0x03);
  if(Set.sopt2 & SO_RC)    // remote connect
    OSTaskCreateExt(Server2Task, (void *)0, &server2Stack[STACKSIZE - 1], 14, SERVER_TASK2_ID, &server2Stack[0], STACKSIZE, (void *)0, 0x03);


/* Old task create
  OSTaskCreate(SerialReceive1Task, (void *)123, &serialReceive1Stack[STACKSIZE - 1], 9);
  OSTaskCreate(SerialReceive2Task, (void *)123, &serialReceive2Stack[STACKSIZE - 1], 10);
  OSTaskCreate(NetReceive1Task, (void *)123, &netReceive1Stack[STACKSIZE - 1], 11);
  OSTaskCreate(NetReceive2Task, (void *)123, &netReceive2Stack[STACKSIZE - 1], 12);


  if(Set.sopt1 & SO_RC)    // remote connect
    OSTaskCreate(Server1Task, (void *)123, &server1Stack[STACKSIZE - 1], 13);
  if(Set.sopt2 & SO_RC)    // remote connect
    OSTaskCreate(Server2Task, (void *)123, &server2Stack[STACKSIZE - 1], 14);
*/


  httpd_init();
  Log (3,"Local IP", I1, &my_local_ip);
  Log (3,"Netmask", I1, &Set.netmask);
  Log (3,"Gateway", I1, &Set.gateway);
  
  telnet_init();
  
  if(id == 0)                 // id 0 = I/O Card connected
    digiod_init();            // Start of digital I/O deamon
    
#ifndef DEBUG                  
  at91watchdogInit();         // Init watchdog
#endif

  while (1) {
    OSTimeDly(OS_TICKS_PER_SEC);
  }
}

int main( void )
{
  at91HWInit();
  
  id = cardId();
  flash = config(&id);
  my_local_ip = Set.local_ip;
  my_serial1 = Set.serial1;
  my_serial2 = Set.serial2;
  Power(0);
  if (flash){
    test = HWtest();
    flashDataWrite(SECTOR_PAR, (char *)&Set, sizeof(Set));
    flashDataWrite(SECTOR_SYS, (char *)&Sets, sizeof(Sets));
  }
  logLevel = Set.log;
  
  OSInit();          
  OSTaskCreate(InitTask, (void *)4711, &initStack[STACKSIZE - 1], 1);
  Power(1);
  OSStart();
  return(0);
}

char cardId(void)
{
  char rc;
  
  at91.pio->PIO_ODR = PIO18;      // Output disable
  at91.pio->PIO_ODR = PIO19;      // Output disable 
  
  rc = (at91.pio->PIO_PDSR & (PIO18 + PIO19)) >> 18;

  at91.pio->PIO_OER = PIO18;      // Output enable
  at91.pio->PIO_OER = PIO19;      // Output enable 

  return rc;
}

void SetAddr (struct sockaddr_in *addr, u32_t ip, u16_t tcp)
{
  bzero(addr, sizeof(struct sockaddr_in));
  addr->sin_family = AF_INET;
  addr->sin_port = htons(tcp);
  addr->sin_addr.s_addr = htonl(ip);
}          

